
from .base import *

from .MPL4 import MPL4
from .MPL5 import MPL5
from .MPL6 import MPL6
from .MPL7 import MPL7
from .MPL8 import MPL8
from .MPL9 import MPL9
from .MPL10 import MPL10
from .MPL11 import MPL11

# Defines the list of datamodels.
datamodel = DAPDataModelList([MPL4, MPL5, MPL6, MPL7, MPL8, MPL9, MPL10, MPL11])

